/* 
** Copyright (C) 1999 by Andreas Junghanns.
**
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation.  This software is provided "as is" without express or
** implied warranty.
*/ 

#include "board.h"

HASHENTRY HashTableNorm[MAX_HASHENTRIES];
HASHENTRY HashTableDead[MAX_HASHENTRIES];
HASHENTRY HashTablePen[MAX_HASHENTRIES];

void InitHashTables() {
	memset(HashTableNorm,0,sizeof(HASHENTRY)*MAX_HASHENTRIES);
	memset(HashTableDead,0,sizeof(HASHENTRY)*MAX_HASHENTRIES);
	memset(HashTablePen,0,sizeof(HASHENTRY)*MAX_HASHENTRIES);
}

void GGStoreHashTable(HASHKEY hashkey)
{
	HASHENTRY *entry = &(IdaInfo->HashTable[hashkey&HASHMASK]);
	
	entry->lock = hashkey;
}

int GGGetHashTable(HASHKEY hashkey)
{
	HASHENTRY *entry = &(IdaInfo->HashTable[hashkey&HASHMASK]);

	if (entry->lock==hashkey)
		return(YES);
	else
		return(NO);
}

void SetPathFlag(MAZE *maze)
{
	HASHKEY    hashkey = maze->hashkey&HASHMASK&~1;
	HASHENTRY *entry   = &(IdaInfo->HashTable[hashkey]);

	if( entry->lock != maze->hashkey ) {
	  hashkey++;
	  entry++;
	}
	if( entry->lock == maze->hashkey )
		entry->pathflag = 1;
}

void UnSetPathFlag(MAZE *maze)
{
	HASHKEY    hashkey = maze->hashkey&HASHMASK&~1;
	HASHENTRY *entry   = &(IdaInfo->HashTable[hashkey]);

	if( entry->lock != maze->hashkey ) {
	  hashkey++;
	  entry++;
	}
	if( entry->lock == maze->hashkey )
		entry->pathflag = 0;
}

void ClearHashTable(MAZE *maze)
/* clear a certain entry in the hashtable */
{
	HASHKEY    hashkey = maze->hashkey&HASHMASK&~1;
	HASHENTRY *entry   = &(IdaInfo->HashTable[hashkey]);

	if( entry->lock == maze->hashkey ) {
	  SR(Debug(6,0,"ClearHashTable: %llx %llx\n", maze->hashkey, hashkey));
	  memcpy( entry, entry + 1, sizeof( HASHENTRY ) );
	} else {
	  hashkey++;
	  if( entry->lock != maze->hashkey ) return;
	  SR(Debug(6,0,"ClearHashTable: %llx %llx\n", maze->hashkey, hashkey));
	  entry = &( IdaInfo->HashTable[ hashkey ] );
	  entry->lock = 0;
	  entry->down = 0;
	}
}

/* entry <,=,> new entry */
int QualityCompare( HASHENTRY *entry, int down )
{
	  if( (int)entry->down > down )
	    return 1;
	  else if( entry->down == down )
	    return 0;
	  else
	    return -1;
}

HASHENTRY *StoreHashTable(MAZE *maze, int down, int min_h,
	int area, int dl, int pen, int pathflag)
{
	HASHKEY    hashkey = maze->hashkey&HASHMASK&~1;
	HASHENTRY *entry   = &(IdaInfo->HashTable[hashkey]);

	if( QualityCompare( entry, down ) > 0 ) {
		  /* even entry is "better": use odd entry */
		  hashkey++;
		  entry++;
	} else {
		  /* odd entry is "better": copy even to odd, use even */
		  memcpy( entry + 1, entry, sizeof( HASHENTRY ) );
	}

	SR(Debug(6,0,"StoreHashTable: %llx %llx down: %i min_h: %i\n",
		maze->hashkey, hashkey, down, min_h));
	if (   entry->lock != 0
	    && entry->lock != maze->hashkey) {
		IdaInfo->tt_cols++;
	}

	entry->lock         = maze->hashkey;
	entry->man          = maze->manpos;
	entry->down         = (short)down;
	entry->min_h        = (short)min_h;
	entry->areasearched = (unsigned short) area;
	entry->dlsearched   = (unsigned short) dl;
	entry->pensearched  = (unsigned short) pen;
	entry->pathflag     = (unsigned short) pathflag;
	entry->goal_sqto    = maze->goal_sqto;

	return(entry);
}

HASHENTRY *GetHashTable(MAZE *maze) {
	HASHKEY hashkey = maze->hashkey&HASHMASK&~1;
	HASHENTRY *entry = &IdaInfo->HashTable[ hashkey ];

	SR(Debug(6,0,"  GetHashTable: %llx %llx down: %i min_h: %i\n",
		maze->hashkey, hashkey, entry->down, entry->min_h));
	IdaInfo->tt_reqs++;
	if (Options.tt==0) return(NULL);

	if( entry->lock != maze->hashkey ) {
		  hashkey++;
		  entry++;
	}

	if ( (entry->lock == maze->hashkey) &&
	     IsBitSetBS( maze->reach, entry->man ) ) {
		IdaInfo->tt_hits++;
		return entry;
	}
	return(NULL);
}

void PSSetPathFlag(MAZE *maze)
{
	HASHKEY    hashkey = maze->hashkey&HASHMASK;
	HASHENTRY *entry   = &(IdaInfo->HashTable[hashkey]);

	entry->pathflag = 1;
}

void PSUnSetPathFlag(MAZE *maze)
{
	HASHKEY    hashkey = maze->hashkey&HASHMASK;
	HASHENTRY *entry   = &(IdaInfo->HashTable[hashkey]);

	entry->pathflag = 0;
}

void PSClearHashTable(MAZE *maze)
/* clear a certain entry in the hashtable */
{
  HASHKEY    hashkey = maze->hashkey&HASHMASK;
  HASHENTRY *entry   = &(IdaInfo->HashTable[hashkey]);
  SR(Debug(6,0,"ClearHashTable: %llx %llx\n", maze->hashkey, hashkey));

  entry->lock = 0;
  entry->down = 0;
}

HASHENTRY *PSStoreHashTable(MAZE *maze, int down, int min_h, int pathflag)
{
	HASHKEY    hashkey = maze->hashkey&HASHMASK;
	HASHENTRY *entry   = &(IdaInfo->HashTable[hashkey]);

	SR(Debug(6,0,"StoreHashTable: %llx %llx down: %i min_h: %i\n",
		maze->hashkey, hashkey, down, min_h));
	if (   entry->lock != 0
	    && entry->lock != maze->hashkey) {
		IdaInfo->tt_cols++;
	}

	entry->lock       = maze->hashkey;
	entry->man        = maze->manpos;
	entry->down       = (short)down;
	entry->min_h      = (short)min_h;
	entry->pathflag   = (unsigned short) pathflag;
	entry->goal_sqto  = maze->goal_sqto;

	return(entry);
}

HASHENTRY *PSGetHashTable(MAZE *maze) {
	HASHKEY hashkey = maze->hashkey&HASHMASK;
	HASHENTRY *entry = &IdaInfo->HashTable[ hashkey ];

	SR(Debug(6,0,"  GetHashTable: %llx %llx down: %i min_h: %i\n",
		maze->hashkey, hashkey, entry->down, entry->min_h));
	IdaInfo->tt_reqs++;
	if (Options.tt==0) return(NULL);

	if ( (entry->lock == maze->hashkey) &&
	     IsBitSetBS( maze->reach, entry->man ) ) {
		IdaInfo->tt_hits++;
		return entry;
	}
	return(NULL);
}

HASHKEY NormHashKey(MAZE *maze) {
/* Creates the Hashkey for maze from scratch, enters it in structure
   and returns it */
	int        i=0;
	HASHKEY    hashkey=RandomTable[0];

	for (i=0; i<maze->number_stones; i++) {
		hashkey ^= RandomTable[maze->stones[i].loc];
	}
	maze->hashkey = hashkey;
	return(hashkey);
}

HASHKEY UpdateHashKey( MAZE *maze, UNMOVE *move) {
/* Updates the HashKey of maze according to UNMOVE, and returns it:
   Assumtion, Moves was made before calling this procedure */

	maze->hashkey ^= RandomTable[move->stonefrom];
	maze->hashkey ^= RandomTable[move->stoneto];
	return(maze->hashkey);
}

unsigned long RandomTable32[1792] = {
0x7462cdf7,0x47b7a5f6,0x1e232b82,0x42db26cd,0x03ce62c9,0x4e8e22d0,0x237f9bfc,0x12a384ef,
0x7a25810b,0x52f3a8da,0x789478a6,0x2fbca201,0x29625d3d,0x08d86e94,0xeb2ee000,0xa87cf97c,
0xeb4987df,0x92990981,0xdb8cfc8a,0x86a5090a,0x42e70d71,0x5751af18,0x1c93a4c4,0x5f1293d7,
0x7e329673,0x33b392e2,0x4eb4fb2e,0x4d32f9a9,0xa4a28765,0xed1437a3,0x7e7a0e48,0x03c860eb,
0xf8be20c7,0xb263bdf9,0x20c97892,0x03e93e1d,0xb2be9f19,0xf74ca19f,0x4fb5008c,0x1c0461bf,
0xa2df5adb,0x97d77815,0xde4838b6,0x9341e7ae,0x16b8e88d,0x02d3d524,0x94821f90,0xb61bb5ac,
0xc3e338af,0xa9775771,0xe8a1bf9a,0x813e63ba,0x2602b7c1,0x425b50a8,0x1195cf54,0x02d78ea7,
0x9ca46e43,0xe246580d,0xfe29513e,0xfc386206,0x0c9120b5,0x0dbeb4ec,0x126933d8,0x42f362bb,
0x856f6f97,0x9116b5e9,0x4e14f1a2,0x111bb16d,0xb120f769,0xc7e25a0f,0xe86a311c,0xdff94570,
0x5db670ab,0x549a12fa,0x2d6164c6,0x61f52aa1,0x22cacfdd,0x454687b4,0xff476b20,0xcf91b5dc,
0x121d657f,0x1687469e,0xa7862eaa,0x9854e26a,0xd89afe11,0xf18581c7,0x7c3845e4,0x79a88577,
0x9bc60213,0xe7b716fd,0xdc8f934e,0xdc57a1b6,0xd9559605,0xf40e9283,0xb224e568,0xa5565f74,
0x8f9bba67,0xc94841d9,0xfed296b2,0xe0e07f42,0x29366bb9,0x271df980,0x2d5c2dac,0x25ef8f5f,
0x193fc27b,0x68114a0a,0x50d6fcd6,0x147ed8f1,0x5475132d,0x39b18644,0xaf4bc2b0,0xa0a9fa0c,
0x3c1d0e4f,0x22dbd0ae,0x1bf349ba,0x1ea47ae5,0x7820e061,0x1d6037c8,0x65b80874,0x51aa7847,
0x1fec51e3,0x4f8a5612,0x568cc15e,0x01d23a99,0xcb68e755,0x8f9b0df3,0x96a422f8,0x8fffe5a4,
0x7d380137,0x48e09e36,0x7cc367c2,0x6261ac0d,0x4f87fc09,0x72c55910,0x1ccbf63c,0x36c3e02f,
0x2290504b,0x2f172d1a,0x96cc00e6,0x82c7dcbe,0x0750b27d,0x2055d0d4,0xf2d02640,0xff6f823c,
0x2c9f331f,0x11bf46be,0xd69e10ca,0xca054bca,0x25395eb1,0x6b7d7d58,0x4e141704,0x02c26717,
0xdd8c5db3,0xcb9910dd,0x3aa7db6e,0x0f0232e9,0x7c3814a5,0x3cba429c,0x0b57ec88,0x65cbd02b,
0x59014407,0x71fcea46,0x6bae64d2,0x5f0f335d,0xa70ea859,0xe51b3b5f,0x09f48acc,0x2d48acff,
0x81bd1a1b,0xfd0b43d5,0x5a0d70f6,0x58be0991,0xafaaadcd,0xdfcb989b,0x044b95d0,0x63d2b193,
0x6e10d3ef,0x2a2aa8ce,0x85c783da,0xc845367a,0x11457901,0x20034ee8,0x800f7194,0xc16aae18,
0x2c5b2583,0x0c87b432,0xa427e17e,0xd84ab8c6,0x86901df5,0xdeada0d3,0xa96d4218,0xd35e3e04,
0x158482d7,0x3765a256,0x44c28de2,0x391516ad,0xd33770a9,0xedf2c3cf,0x4d24eb5c,0x0602f5cf,
0x082b1feb,0x19b2f73a,0xc48e4d06,0x8892741e,0x21c4051d,0x570e49f4,0x9ed91160,0xb16e5e9c,
0xfcfef0bf,0x8c290921,0x808ca2ea,0x8a4e452a,0x3a562f51,0x1ce2ac78,0x205f1824,0x408838b7,
0xbb4da953,0xfdaa5abd,0xb133d38e,0x9bd78876,0xa1020345,0xf2b1b843,0x8a5723a8,0xed291034,
0xc236bda7,0x98cc3999,0xd7e6e2f2,0xfbdfaa02,0x5d1f54f9,0x4d8fbfc0,0x0c2417ec,0x2f37ba9f,
0x96bf61bb,0xe9e521b5,0x7aa79516,0x0d63aa31,0xda35b86d,0x969b877b,0x569b98f0,0x4c024d33,
0x1416898f,0x473d30ee,0x79e46dfa,0x596c8825,0x5ac481a1,0x0ecc9608,0xd2440ab4,0xe23fe478,
0x5ad8e923,0x34f9c252,0xf832b19e,0x97a73c26,0x6b22c495,0x3e8efc4c,0x6bc49138,0x1480599b,
0x9be4f477,0xd7bfa989,0x49406402,0x139df14d,0x822b5549,0xaa82b0af,0x94f5107c,0xa7140490,
0xde0edf8b,0xe84a8ea5,0xa9ae4926,0xfd829b7e,0xaf28c7bd,0xed880ceb,0x7ed22c80,0x7d19b503,
0xfd549e5f,0xa269a901,0x9107e50a,0xcbc7ce8a,0x5ab16ff1,0x7d320b98,0x0a254944,0x1521fa57,
0xed51e4f3,0xb7a2f49d,0xf0cb7bae,0xcbfcd3d6,0xa0c361e5,0xae4a8323,0x76ea8ac8,0x2b72ff6b,
0xb7342747,0xf5dbad79,0xfa001112,0xb5451762,0x43bc7199,0x54a6eae0,0x8b6ad50c,0xbb1b47c0,
0x718e995b,0x4acbb06a,0x78db6936,0x4257bad1,0x900e330d,0xcbb6465b,0xc733cc10,0x9773272c,
0xff7e2f2f,0x9f2496f1,0x9722081a,0xf0de493a,0xf40dfa41,0xa5bbf2d7,0xef1dd3d4,0xa0ad2ad8,
0x5acd9cc3,0x75288072,0x766531be,0x72a4b079,0xec7cdb35,0xaddd3693,0xd8981058,0xe0bc1ec4,
0xe2015617,0xeac74569,0x0b8eea22,0x6c643bed,0x863ba9e9,0xd7626d8f,0xb884659c,0xe9590ef0,
0x7ee38f2b,0x78669b7a,0xbe35f546,0xf3b052de,0x748afa5d,0x679acc34,0xfb6b77a0,0xb52f475c,
0x5a603bff,0x7bc5671e,0xd247d72a,0x8509e7ea,0x16db2091,0x12f39ab8,0x166eaa64,0x7eb7abf7,
0x00611093,0x18c52182,0x4ec8d3ce,0x4b7a50c9,0x14943085,0x49f7e1fc,0x1cde21e8,0x5547ff0b,
0x1f6180e7,0x5a968ea6,0x738fef32,0x6b46eb3d,0x9245fe39,0x8e4db9ff,0x3300c22c,0x2377a5df,
0x94c2c0fb,0xe9b0cd75,0x6722ed56,0x43823b71,0xe8cc1dad,0xf413d53b,0xe54a2f30,0x99e5ab8c,
0x1ce7c4cf,0x69cd512e,0xbd8a523a,0x9fbdaa9a,0x5181e2e1,0x1ef1b448,0xba8cccf4,0xf08a8138,
0xe5a14063,0xbaa4116d,0x1ff961de,0x31810d19,0x8fe661d5,0xadea3973,0x20cfbf78,0x209458db,
0x1c51a7b7,0x0416ceb6,0x70ea2042,0x39cff68d,0xcb5c6e89,0xdd09fa6f,0x5c98eabc,0x425bd6af,
0xfeb12ecb,0xa8f18a65,0xd4e95166,0xc1339a3e,0x250a9cfd,0x577ed554,0x2264f2c0,0x525eac43,
0x6359c99f,0x072c273e,0x2784794a,0x4f536eb5,0xbd0f4131,0xef50a627,0xe5c13b84,0xe88eb268,
0xc6032c33,0xae2a185d,0x35c1dbee,0x3375e569,0xf4846f25,0xdde288e3,0xd3b9e908,0xee021154,
0x9e1eca87,0x9bad8539,0xd7487d52,0xe5d3a222,0x238bfad9,0x1239d120,0x0a2fdf4c,0x1c18837f,
0xa003d89b,0xc6129b55,0x3f622176,0x18cb2c11,0xc227784d,0x98ac341b,0x483ac250,0x4d52c013,
0x83e34a6f,0x9f2516b1,0x440b4c5a,0x50366405,0xf7983b81,0xdaa27497,0x208ef614,0x56501867,
0x7dbbd403,0x69dc0cb2,0xfd2541fe,0xb42a2646,0x8efb5875,0xdf900c53,0x234f9e98,0x4819c07b,
0x4f3de957,0x06a292d6,0x90fe0662,0x91b6ded2,0x2b6da329,0x080ea8b0,0x71c89fdc,0x7b32ac4f,
0xf9dfbe6b,0xe50b0045,0xc3a45d86,0x8024719e,0x8ff3af9d,0xb0d3f18b,0x8aac9de0,0xf594701c,
0xf531473f,0xbb6d68a1,0x1537cb6a,0x32b83555,0x6925d1d1,0x54ed48f8,0xcd24fca4,0xea8920c8,
0x264037d3,0x1dd75dc2,0x49ee940e,0x30ddea09,0xe5681dc5,0x8dd1c3c3,0xdd81e028,0xc0c331b4,
0x955c0427,0xc0dfe919,0xeb85bb72,0xf82cbf82,0x7de66779,0x4fc58c40,0x6ea22c6c,0x1bef511f,
0x8019e03b,0xbd11b935,0x37050596,0x061a8cb1,0xa13842ed,0xd67762fb,0x649d8570,0x02de1bb3,
0xcf88c00f,0xaf33ce91,0x2326f67a,0x5e65e2a5,0x5c310421,0x620f9288,0xb3124f34,0x87f55df8,
0x490557a3,0x72f0dad2,0x9b22d21e,0xdb74e9a6,0xf5e7bf15,0xf4c6af33,0x2681adb8,0x217c181b,
0xa75e1af7,0xc55bf909,0x17129c82,0x7837bbcd,0x8bdf47c9,0xb890842f,0xbd9d84fc,0x83ac8e10,
0xafe73e0b,0x889dc625,0xaa5d19a6,0x8e9ad8fe,0x4e4e323d,0x27aa7794,0xdab87900,0xba609c7c,
0x0246b4df,0x04c0b77e,0x8cd7cd8a,0xdd73940a,0x2152d271,0x6a356818,0x0789edc4,0x4af060d7,
0x15403373,0x731183e2,0xcd88fc2e,0xbb65a156,0xe2ab3c65,0xc1cdcea3,0x37be0748,0x16ac9deb,
0x68412dc7,0x0ec76306,0xfe85a992,0x9f5c6ce2,0x83354419,0x9022889f,0x2647a98c,0x2c240ebf,
0xc0ecd7db,0xf5662715,0x680399b6,0x0b585d51,0xf0367d8d,0xde6d61db,0x1fe47890,0x47646753,
0xf4f825af,0xaf96d671,0x3b99509a,0x0638d145,0xa1fc3cc1,0x8a703657,0x8016d854,0xab32e458,
0x0565cb43,0x3ce258f2,0xe5a6123e,0xa5773d06,0x564395b5,0x4c79ddec,0xa32decd8,0xeb9ca044,
0xeb5a3c97,0xb61cd4e9,0x6c6de2a2,0x4203c66d,0x924d5c69,0x835f810f,0x03839a1c,0x1ce6278f,
0x72afadab,0x6b9e23fa,0x4a1d85c6,0x71512fa1,0x576224dd,0x260210b4,0x2ed08420,0x32a22723,
0xdb2a127f,0xdb817861,0x4e5c7faa,0x18381295,0xd4a24311,0xf8f048c7,0x8ae80ee4,0xd9fa2d88,
0x65ab1f13,0x13cc5a02,0xe6c5144e,0xd46cbcb6,0x1c19cb05,0x7631567c,0x6afa5e68,0x2df55d8b,
0xf14e4767,0xfcefa0d9,0xcb9047b2,0xb2faaa42,0x7d4890b9,0x64099280,0x56c056ac,0x71debc5f,
0x1044bf7b,0x36381b0a,0xeb93ddd6,0x9b93620e,0x24fe282d,0x0a79cf44,0x13cf9bb0,0x378da2f3,
0xc4897b4f,0xfd862e51,0x60fe5aba,0x3f172fe5,0x6589e561,0x2c6630c8,0x668c9174,0x68bf8547,
0x3f052ee3,0x6ff88712,0xa16b025e,0xb7492066,0xb966dc55,0xf5c664f3,0x8aba5bf8,0x9b8868a4,
0x432a4e37,0x6a27ff36,0xff3fd8c2,0xcdeabef2,0xee6fe109,0x90d64def,0xf71adf3c,0xafed32d0,
0xcfa10d4b,0xbd0f41e5,0x646ba1e6,0x1987a841,0xc217877d,0x84c5262b,0x742cbf40,0x5c1bdac3,
0xbf3b601f,0xf36bf841,0xb1bfe1ca,0xf6dcd6ca,0x70fc23b1,0x0e043658,0x385f6004,0x30df3417,
0x12e8fab3,0x0e4fe022,0xc9a0dc6e,0xa14f6816,0x398bc9a5,0x7633ab9c,0x04bce588,0x76bf0d2b,
0x9bab5107,0xa83cf4b9,0xb23795d2,0x8e9f77a2,0x54bc4d59,0x49d1dda0,0x2a1233cc,0x2e4759ff,
0xbc79971b,0xc8eef2d5,0xc6fdd1f6,0xf6c0b16e,0x177b42cd,0x0b463064,0x3caeeed0,0x4c01ce93,
0x6e8cc0ef,0x15c629ce,0x6b1414da,0x7c68fe85,0xe819fe01,0x80e53817,0x16f57a94,0x4c09dee7,
0xe80b8283,0xff849acd,0xb329a27e,0xfa0293c6,0x14b592f5,0x5680882c,0x18f2fb18,0x3160befb,
0xc0b64fd7,0xa2c57ca9,0x04007ee2,0x2ad42bad,0xeb1ad5a9,0xed6ceacf,0x0ba9545c,0x250162cf,
0x64135ceb,0x14a2083a,0xe85f6e06,0xc00f6f1e,0x64165a1d,0x445cd2f4,0x4df72a60,0x28b47e63,
0x99ea9dbf,0xce36c821,0x7c77f3ea,0x66b5afd5,0x0b187451,0x0d9ae578,0x02d9e124,0x11a485b7,
0x8f61c653,0xc31be9bd,0x1e8e548e,0x12da5c89,0x5f993845,0x064130bc,0x178f9ca8,0x3eb1accb,
0xb0f84aa7,0xab589899,0xa87193f2,0x91e2d502,0xaba879f9,0x9541a73f,0xd32540ec,0xc17a1860,
0x55335ebb,0x4a6caf4a,0x31d97616,0x2db86f31,0xda51cd6d,0xc9003e7b,0xcc1071f0,0xa69715cc,
0x84a1f68f,0xbae9ce11,0x339c7efa,0x2d963d25,0xaa5486a1,0xedca70f7,0x83c593b4,0xcd2bd778,
0x5ce0c623,0x2eb2f352,0x8797f29e,0x9abb9726,0xa9a3b995,0xd0a95ab3,0x24ffca38,0x76c6d69b,
0x2e564177,0x70e2b776,0x3129d502,0x0aa8864d,0x5d4a3a49,0x0664a850,0x3676f97c,0x46d9e86f,
0x78ae9c8b,0x43fa025a,0x6d22ea26,0x7673e981,0x99469cbd,0xf59003eb,0x51e3c580,0x5b141203,
0xb417cb5f,0xeb19e801,0xadfcb60a,0xe1a8598a,0x2a5f34f1,0x405bc498,0x21b99244,0x0a7dc757,
0xf21d81f3,0xba2f039d,0x66897cae,0x38f89129,0xc65a16e5,0xa59e1a23,0x6f7483c8,0x2f753c6b,
0x1edd3447,0x4d857386,0xb4564212,0xa05cc262,0x7e251699,0x3c5703e0,0xd90d7e0c,0xe43d9ac0,
0x7efa165b,0x7993016a,0xd660ca36,0xbe40002e,0x1331c80d,0x0cae82a4,0x90aa2510,0xef950a2c,
0xe8591c2f,0xa8ce15f1,0x286d991a,0x3e06ebc5,0x3c897f41,0x363e8628,0x8474dcd4,0x92b99dd8,
0xe3ecf9c3,0xf918ce8d,0x4eedf2be,0x7573d579,0x64195035,0x7ec3f26c,0xb782c958,0xf3ae21c4,
0x77722317,0x7ee89b96,0x9b39db22,0xd405af12,0xacd60ee9,0xe9d9948f,0xb27bce9c,0x833ba1f0,
0x08facc2b,0x0040ac7a,0x671a1646,0x0af9b221,0x1bcc4f5d,0x427c5534,0xc27a90a0,0xee1d6a5c,
0x6cb2e8ff,0x42b2a81e,0x2c46282a,0x19710d15,0xec146591,0xf3312c47,0x98b47364,0x866d0708,
0x2ba42d93,0x1b5e9282,0x528854ce,0x7df335c9,0x942a6585,0xed623503,0xb17b9ae8,0xb14e43f4,
0x232a0de7,0x22252fa6,0xdee7a032,0xa1a53fc2,0xda462339,0x9fdc20ff,0x8792eb2c,0xa4db2d20,
0xe5b5bdfb,0xe633fc75,0x1d39ce56,0x713e0071,0x578332ad,0x775a73c4,0x5dc40830,0x7daff173,
0x7dd231cf,0x1ee1522e,0x47df633a,0x15230a65,0x3d38e7e1,0x67bdad48,0x73b955f4,0x67888bc7,
0xe1181d63,0x9e9fe06d,0x5a63a2de,0x0173b219,0x31ee56d5,0x1ad06f8c,0xc18bf878,0xda562a24,
0x10c1f4b7,0x6d142fb6,0x4ff89142,0x65318b8d,0x01b65389,0x39605e90,0x9351d3bc,0xd6173c50,
0x0fafebcb,0x44be069a,0x0642f266,0x0769eac1,0x809b71fd,0xb27621ab,0x1f298bc0,0x47c80943,
0xfb3bf69f,0xac0917c1,0x21ce4a4a,0x3d69e3b5,0xcfb40631,0xfa83ed27,0xd6668484,0xa6f3e568,
0x261dc933,0x7bd4d8a2,0x9344dcee,0xb54db596,0xc2662425,0xc5031fe3,0xa0a4e208,0x81f0d454,
0x2aa6d787,0x3e4e9bc6,0x56abae52,0x7eabb2dd,0x492b9fd9,0x23acea20,0x5953884c,0x2dd5307f,
0x84ae559b,0xb3a04a55,0x81ac8276,0xdbe58eee,0x1ea60d4d,0x4e0b94e4,0x08161b50,0x09dfdd13,
0xf91d376f,0x851395b1,0x9123dd5a,0xe9ad66fa,0x72cec081,0x4f3b0468,0xc952ff14,0xd63d5a98,
0xb64a3103,0x9d9a424d,0x9c3102fe,0xefd40146,0x620ecd75,0x45841cac,0xd99d5798,0xad894284,
0x566db657,0x2d2d73d6,0x6d17f762,0x08b6362d,0x4dc30829,0x139a81b0,0x89fb08dc,0xb290e6b0,
0xf4e5fb6b,0xa745ef45,0xba2f7e86,0xce536c9e,0x9f2c049d,0x875f688b,0x2570b6e0,0x536c6ce3,
0x20e2f43f,0x6d7ad85e,0x96431c6a,0x9455d5aa,0xfe8616d1,0x97147e07,0x705fc5a4,0x3f112c37,
0xaed254d3,0x8e84313d,0x1493150e,0x441dcf09,0x2a5152c5,0x4b87253c,0x5b7e5928,0x7b358b4b,
0xcd1b9127,0xe6364819,0x5aaa6c72,0x50fe157d,0xa7658c79,0xff85dabf,0x2149556c,0x60fb7e1f,
0x0d3bdd3b,0x7e9617ca,0x1cf2e696,0x7e3d51b1,0xca0657ed,0xac3619fb,0xa9e85e70,0x865d474c,
0x939a2d0f,0xa9e4cd91,0xf5e7077a,0xdf02685a,0x68130921,0x413e8b88,0x71a9d834,0x581caf07,
0x44eb34a3,0x7a400bd2,0x4a90131e,0x6484bb59,0x3d16b415,0x2ce6f9cc,0x16fee6b8,0x7e60951b,
0x8b0d67f7,0xc3039809,0xb2ec0d82,0x9b0faf32,0x42f02cc9,0x745cd4d0,0xb7876dfc,0xfd80a110,
0x7164fb0b,0x4f7ccada,0x4f8bbaa6,0x0ba9ac01,0xe436073d,0x83377f6b,0xe8ce1200,0x98883f7c,
0x24c7e1df,0x4006787e,0xff769e8a,0xfe661f0a,0x4aa29771,0x21a52118,0x2dfa36c4,0x5dca2dd7,
0xa9a9d073,0xad648b1d,0x4a30fd2e,0x4b1dc3a9,0xf55ff165,0xd7bb65a3,0x0f4e0048,0x5fccdaeb,
0xb7e83ac7,0xc7a17bf9,0x982dda92,0x9e4617e2,0x794fe919,0x1c139060,0xf744528c,0xa2404440,
0x2dc654db,0x69b729ea,0x2362fab6,0x5f8ea251,0x0108128d,0x379d6724,0x4a52d190,0x6da08453,
0x751912af,0x2735aa8e,0x392ee19a,0x108c0645,0x4f0dc1c1,0x3e5042a8,0xb247e154,0xcd415758,
0xbe632843,0x9ac8f60d,0x1138d33e,0x7d65e7f9,0xc35e0ab5,0x94fef913,0x985ca5d8,0x8ef0a344,
0x3a490997,0x26490c16,0xf2f2d3a2,0xbbb82492,0xf87dc169,0xa8d0a80f,0x47ef031c,0x0041948f,
0x65c4eaab,0x084e34fa,0xdf3da6c6,0xacb6cb5e,0x707179dd,0x7f0999b4,0xad599d20,0xed6dfbdc,
0x4fbabf7f,0x0861c89e,0xab86d0aa,0xa75ef86a,0x02218811,0x4630f038,0xec6dd7e4,0xfca0e088,
0x914c3c13,0xbe8434fd,0xb130954e,0x8365d7b6,0xde0e0005,0xdec2c083,0xa55dd768,0xde82e574,
0x7b34d467,0x29d50026,0x4d83f8b2,0x39472abd,0xc53eb5b9,0xb7fed47f,0xcc787fac,0xd6b616a0,
0xe3d5bc7b,0xecf513f5,0xfc24bed6,0x9c099d0e,0xe61f3d2d,0xeb71e7bb,0x045f74b0,0x62813ff3,
0x42b9e84f,0x7103d2ae,0xcb556bba,0xc39a1b1a,0xa092ea61,0x8d07d637,0xcbd71a74,0xea2f6db8,
0x589a0be3,0x2b92b812,0x93df435e,0xea487b66,0x1c98d155,0x05da440c,0x6c3c94f8,0x4e2b145b,
0x72d89b37,0x7edb6036,0xbf2649c2,0x92db29f2,0x365bc609,0x639a0b10,0xb8b9c83c,0xe62245d0,
0x30bdca4b,0x2e764f1a,0xe43d42e6,0xfd8cd2be,0x86765c7d,0xef941d2b,0x1f775840,0x3f6337c3,
0xe55b8d1f,0xf0ab3741,0x9f2bb2ca,0xb9d861ca,0x1056e8b1,0x1916ef58,0x135aa904,0x19f80117,
0x518197b3,0x2c64d122,0xb3addd6e,0xce850316,0x4c077ea5,0x7279149c,0x6571de88,0x11ee4a2b,
0x2e315e07,0x09f52c46,0xede4c6d2,0xa3f222a2,0x0451f259,0x01caf6a0,0x77eddccc,0x64c206ff,
0xfd02141b,0xb426a1d5,0xa07032f6,0x9ca36c6e,0xa8a7d7cd,0x8e5c069b,0x289047d0,0x30eceb93,
0x5e8cadef,0x424daace,0x7016a5da,0x77f33385,0x0a9e8301,0x37be40e8,0x6f978394,0x797a6be7,
0x2977df83,0x219b1632,0xf23f637e,0xdb9e6ec6,0x34cb07f5,0x6f7ab12c,0x0406b418,0x155bbbfb,
0xfc641cd7,0x97849ba9,0xe4126fe2,0xde10bf52,0x8a423aa9,0x92db11cf,0xc547bd5c,0xc1843030,
0x6ac799eb,0x393d193a,0x9f948f06,0xb0f06a1e,0x969caf1d,0xd208a40b,0x3a314360,0x34935b63,
0x295a4abf,0x29a778de,0xbb7544ea,0xe46a5b2a,0x8eaeb951,0xf720e187,0xfdaeaa24,0xd8432d48,
0x7c51e353,0x4d9e8742,0x4b3ed58e,0x7ea84189,0xa3d06d45,0xbbffe643,0xd15215a8,0xa1379634,
0xa6a5d7a7,0xff78f799,0x25a844f2,0x4a75fffd,0xb6199ef9,0x8d070e3f,0x12ee69ec,0x4d50149f,
0x38735bbb,0x316a804a,0xb94f5716,0x8e56cbce,0xd409e26d,0xbe18f57b,0x74654af0,0x0d8b8733,
0x2ff1638f,0x35db32ee,0xb4608ffa,0xed640dda,0xb3788ba1,0xeed577f7,0x73391cb4,0x71e43587,
0x3064a323,0x67982452,0x090f339e,0x5c4c0dd9,0xa8ccae95,0xfa15b1b3,0xea850338,0xddb6ac64,
0x17bb8e77,0x10f11876,0xc35d4602,0xdef0e4b2,0x80f11f49,0x8ca7feaf,0x68c6e27c,0x7943d56f,
0x970a598b,0x94156ca5,0x03498b26,0x2b066e81,0x541871bd,0x40b40514,0x4bef5e80,0x14ca6f03,
0xea56f85f,0xc7de2701,0xae45870a,0xd9ace48a,0x0e90f9f1,0x30117d98,0x3241db44,0x4ed59457,
0x04851ef3,0x2b70ed62,0xddbb7dae,0xcaf609d6,0xf698cbe5,0xd625b123,0x0d4a7cc8,0x11b3796b,
0xc27a4147,0xbead6b79,0x99507312,0xff186d62,0x8371bb99,0x8eece31f,0x3968270c,0x3c1c123f,
0xe731935b,0xd2f9ad95,0x7b182b36,0x5ec444d1,0xaeb55d0d,0x9ba0b45b,0xb81e7e10,0xf6faed2c,
0x6438092f,0x72746b0e,0x3edb2a1a,0x57c820c5,0x90690441,0xd03b00d7,0xc483e5d4,0xd0ca10d8,
0x9cc856c3,0x962e1d8d,0xdb48b3be,0xd3a10586,0x0e01c535,0x13311b6c,0x7a7b8258,0x2c9bdb3b,
0x77e6f017,0x32047c96,0x34b0cc22,0x24ec65ed,0xc25873e9,0xde44bb8f,0x50cd379c,0x715dcb0f,
0x283e092b,0x55c6bd7a,0x70743746,0x443fb721,0xf2eda45d,0xe25621cb,0x0879a9a0,0x5eb072a3,
0xdd0195ff,0xd87416e1,0xa912792a,0xd037fdea,0xb3c5aa91,0xaac9f347,0x75563c64,0x4d6a45f7,
0xc2234a93,0xe7dbfc7d,0xc37dd5ce,0x9e77e536,0xd4b49a85,0xdbf04c03,0x5fa313e8,0x1057790b,
0xe7169ae7,0xf3e02f59,0xda6b5132,0x88356ac2,0x4e2e4839,0x7da17800,0x3653142c,0x264dffdf,
0x23a4bafb,0x73f4d48a,0x4d84af56,0x3695c571,0x1b1247ad,0x2414bcc4,0x9babe130,0xeffe718c,
0x8c809ecf,0xd51eacd1,0x7658743a,0x0edfbf65,0xaf97ece1,0x8fea59b7,0x2d1fdef4,0x7d9798c7,
0x8e8afa63,0x9f6faf6d,0xaee1e3de,0xc07da8e6,0x20324bd5,0x4d57188c,0xb9f23178,0xb804ad24,
0xba7641b7,0xee826f49,0x20990242,0x63ef208d,0x73343889,0x4fd6b790,0x8252bcbc,0x960e4f50,
0xf49aa8cb,0xade66865,0x334e9366,0x04a36fc1,0x39c446fd,0x05e0e754,0x04f624c0,0x6ced6643,
0xcb9a239f,0x8f5256c1,0x4eec1b4a,0x7d5c58b5,0xf1d0cb31,0x952b3427,0xffdbcd84,0x825d1868,
0x73546633,0x31ffc9a2,0x51bbddee,0x4d0aaf69,0x916fd925,0xe157b6e3,0xf8dfdb08,0xe3a39754,
0xe1c2e487,0xd3494339,0x34f2df52,0x238707dd,0xcaef44d9,0xb9d3fcdf,0x720b314c,0x1d0ddd7f,
0x7e14d29b,0x077e06aa,0xf27ae376,0xbefa49ee,0xce7ca24d,0xc7f0a21b,0x4c1f7450,0x6b28fa13,
0xcbe3246f,0xe21614b1,0xfef66e5a,0x84b531fa,0x0a894581,0x5aa47d68,0xad450814,0xbacecd98,
0xd6948e03,0xafe4914d,0x5f12c3fe,0x7c9e23b9,0x46fe4275,0x566445ac,0xcb571098,0x9bf04584,
0x9f918357,0xfedbab29,0x9d1be862,0xef80b4d2,0x069c6d29,0x01325ab0,0x973f71dc,0xbdd879b0,
0xa148386b,0xf7d4de45,0x8d1e9f86,0xede6679e,0x52e4599d,0x26612074,0xde30cfe0,0x89d6b61c,
0x7a10a13f,0x184f195e,0x3f346d6a,0x50781f55,0xf1de5bd1,0xf08a4507,0xd2ca8ea4,0xaa5886c8,
0xe5e071d3,0x9cb3c03d,0x448d960e,0x7c79b409,0x141a87c5,0x71ac0e3c,0x8008d228,0xac95b7b4,
0xeec71e27,0x8520a719,0x407b1d72,0x7484ea7d,0xa6c4b179,0xdbc541bf,0xb4187e6c,0xb27c54e0,
0x3799da3b,0x34e9e8ca,0xfa32c796,0xfe03e94e,0xdaa86ced,0xbca8d0fb,0x8b733770,0xc0dcaa4c,
0x21ef9a0f,0x1656336e,0xc61d187a,0xa38eb35a,0x22890e21,0x23f98488,0x66bb6134,0x0a2abc07,
0x7c0d11a3,0x08bb3cd2,0x5213541e,0x069a6059,0xc8ada915,0xc39f5d33,0x83ca1fb8,0xb37eede4,
0x06f0b4f7,0x5ec0c8f6,0x873b7e82,0xa6fb1a32,0x264911c9,0x65562dd0,0x453d56fc,0x75274bef,
0xf0aeb80b,0x94bca425,0xfa6c5ba6,0xb175cefe,0xa3d5dc3d,0xe7cd766b,0xaa41ab00,0xf2f3e27c,
0x46c50edf,0x2f38397e,0xe3616f8a,0x997caa0a,0xecd65c71,0xf25f25e7,0x38787fc4,0x679ffad7,
0x8c4f6d73,0x9dae9a1d,0x536cfe2e,0x30bd28a9,0xb32ca665,0x9edcfca3,0x076bf948,0x2f2917eb,
0x058b47c7,0x5a61a506,0xa0800b92,0xa8d3c2e2,0x398a8e19,0x1d55a960,0xf28efb8c,0xff289740,
0x0c5bd1db,0x15807aea,0xded45bb6,0xc69f18ae,0x0f39a78d,0x54f43024,0x18df2a90,0x0c98a153,
0x093dffaf,0x6aee2b8e,0xc6ee729a,0x9244c4ba,0x15bf46c1,0x2c9cbba8,0xae28ea54,0x9353ca58,
0x579c8543,0x26b7baf2,0x5e9d943e,0x3c5f0cf9,0xfdb07fb5,0x86abd013,0x527b5ed8,0x28f759bb,
0xa833d697,0xebe512e9,0x4f63c4a2,0x67e7f06d,0xcd4e2669,0xa835cf0f,0xfe9c6c1c,0xe5e6fe70,
0x5b2627ab,0x3aaa45fa,0xd7c1c7c6,0xa822c65e,0xe920cedd,0x9fa2dd4b,0x62dab620,0x203de123,
0x6a4f6c7f,0x5231099e,0x6cbb21aa,0x78e5fc95,0xc900cd11,0xa421d6c7,0x04a5a0e4,0x61b46c77,
0xe6a95913,0xcea8c3fd,0x5b6e164e,0x66bd0d49,0x691a3505,0x1d15287c,0x624b5068,0x0540d78b,
0x2af76167,0x3b05a126,0x8a99a9b2,0xcc1b0042,0xb918dab9,0x9cfb3b7f,0x0b44a8ac,0x1c31165f,
0xc462b97b,0xf17642f5,0x2ed99fd6,0x631c27f1,0x83d8522d,0x98119ebb,0x9b3b4db0,0xcfcf230c,
0xca2e554f,0x81862c51,0x91f87cba,0x996f661a,0xc933ef61,0xfee9dd37,0x0453a374,0x28dd9f47,
0xbe2ae8e3,0xeda716ed,0x6c67845e,0x38d42999,0x9cdec655,0x8cb912f3,0x112acdf8,0x7d1a915b,
0x84c2e837,0xe9053ec9,0xaefabac2,0x9c6f94f2,0xf16bab09,0x9fe99bef,0xe520b13c,0x9bdb58d0,
0xedf6874b,0xfe581fe5,0x6a74e3e6,0x07fab241,0x0f8d317d,0x6fe8ebd4,0x18abf140,0x7a6694c3,
0xc0ffba1f,0xd5fe7641,0x9b7383ca,0xc2f7ecca,0x0c25adb1,0x1cb5a858,0x6ce5f204,0x0e0cce17,
0x7c7634b3,0x1ea5c222,0x2a8ede6e,0x576161e9,0x2f9f33a5,0x418a7d9c,0x3baed788,0x0759872b,
0x86536b07,0xe76cb2b9,0xdf85f7d2,0x90e8cda2,0xe2079759,0xad2ff05f,0x7ba385cc,0x20b8b3ff,
0x5f56911b,0x514daf2a,0x4e9893f6,0x3715d891,0xcf2c6ccd,0x9cb23d9b,0x009fa0d0,0x62940893,
0xd6109aef,0xc03ed431,0x86c336da,0x85a6977a,0x8ceb0801,0xa6124617,0x6ae58c94,0x16e6f8e7,
0xd6a03c83,0x801938cd,0x6569247e,0x52e1b639,0x1e947cf5,0x7c40da2c,0xb4866d18,0xd76d4704,
0x430de9d7,0x69284556,0x2c0860e2,0x6e6655ad,0xee919fa9,0xce3d38cf,0x9d20265c,0xdd8dc330,
0x66b7d6eb,0x17842a3a,0x96bdb006,0x8b35651e,0xaf67041d,0xdc221b0b,0xc7775c60,0xbdd1c79c,
0xdb4df7bf,0x948e4621,0x738095ea,0x645199d5,0x3aa8fe51,0x1eaf5778,0x57b37324,0x52d11fb7,
0xe8de0053,0xbb7b07bd,0x4941568e,0x77922689,0x40f7a245,0x568b02bc,0x79a28ea8,0x431b26cb,
0x8b6764a7,0x852d5699,0xe88cf5f2,0x96d52b02,0x3aeac3f9,0x763f8ac0,0xd04792ec,0xd469be60,
0xd07f58bb,0xa4ebaeb5,0xef593816,0xd6ca06ce,0x811df76d,0xe5e5ac7b,0xd3c223f0,0xc795dbcc,
0x8f84d08f,0xd673cc11,0xc4a0a0fa,0xa78258da,0x8f4090a1,0xe4547ef7,0xa226a5b4,0xd50fbd78,
0x1ba48023,0x6fa95552,0x869c749e,0xe2904d26,0x1f35a395,0x5249f74c,0x96d43c38,0x98f82f64};

/* HASHKEY *RandomTable = (HASHKEY*) RandomTable32; */

HASHKEY RandomTable[ 896 ];

/* Kludge to get an array of aligned long longs */
void InitRandom()
{
  int i;

  for( i = 0; i < 1972; i++ )
    ((unsigned long *)RandomTable)[ i ] = RandomTable32[ i ];
}
